"""
Pypodo scripts
"""

import os
import re
import sys
import time
import base64
import hashlib
from pathlib import Path
from shutil import copyfile
from datetime import datetime
from datetime import date
import configparser
from termcolor import colored
from cryptography.fernet import Fernet
from cryptography.fernet import InvalidToken

STR_PATH_HOME__TODORC_ = str(Path.home()) + "/.todo.rc"
STR_PATH_HOME__TODOCRYPT_ = str(Path.home()) + "/.todo.crypt"
STR_PATH_HOME__TODODECRYPT_ = str(Path.home()) + "/.todo.decrypt"
REGEX_INDEX = "^\\d+$"
REGEX_SPACE_OR_ENDLINE = "( |$)"
RED = "33m#"
YELLOW = "31m#"

# entrypoint
def pypodo(openfile=open):
    """
    Entrypoint
    """
    if len(sys.argv) == 1:
        helppypodo()
    elif sys.argv[1] == "list":
        decrypt(True, "", open)
        listtask(openfile)
        crypt(True, "")
    elif sys.argv[1] == "add":
        decrypt(True, "", open)
        add(openfile)
        crypt(True, "")
    elif sys.argv[1] == "del":
        decrypt(True, "", open)
        delete(openfile)
        crypt(True, "")
    elif sys.argv[1] == "sort":
        decrypt(True, "", open)
        sort(openfile)
        crypt(True, "")
    elif sys.argv[1] == "help":
        helppypodo()
    elif sys.argv[1] == "untag":
        decrypt(True, "", open)
        untag(openfile)
        crypt(True, "")
    elif sys.argv[1] == "tag":
        decrypt(True, "", open)
        tag(openfile)
        crypt(True, "")
    elif sys.argv[1] == "backup":
        decrypt(True, "", open)
        backup(openfile)
        crypt(True, "")
    elif sys.argv[1] == "find":
        decrypt(True, "", open)
        find(openfile)
        crypt(True, "")
    elif sys.argv[1] == "crypt":
        if len(sys.argv) > 2:
            crypt(False, sys.argv[2])
        else:
            crypt(False, "")
    elif sys.argv[1] == "decrypt":
        if len(sys.argv) > 2:
            decrypt(False, sys.argv[2], open)
        else:
            decrypt(False, "", open)
    else:
        helppypodo()


# primary functions
def helppypodo():
    """
    Display help message
    """
    help_txt = """\

    NAME
        pypodo

    SYNOPSIS
        pypodo is a todolist tool which works with a .todo file positionned the root of the home directory
        pypodo [MODE] [PARAMETERS]...

        help    : display this help
                  -> pypodo help
        list    : display the todolist filtered on the tasks tagged with [PARAMETER]...
                  -> pypodo list #print the full todolist
                  -> pypodo list linux #display the todolist filtered on the task tagged with linux
                  -> pypodo list linux urgent #display the todolist filtered on the task tagged with linux and urgent
        add     : add task(s) in [PARAMETER]... to the todolist with an index autogenerated
                  -> pypodo add "my first task #linux" "my second task" #add the task "my first task #linux" (tag with linux) and the task "my second task" to the todolist
        del     : delete task(s) identified with the index equals to [PARAMETER]... from the todolist
                  -> pypodo del 3 4 #delete the tasks identified by index equals 3 and 4
        tag     : add the tag [PARAMETER[1]] to the task the task identified with the index equals to [PARAMETER[2]]... 
                  or without parameters, display the tags of the todolist
                  -> pypodo tag linux 3 4 #add the tag linux to the task identified with the index equals to 3 and 4
                  -> pypodo tag #display the tags of the todolist
        untag   : delete the tag identidied by [PARAMETER1] in the task definied by the [PARAMETER2]... 
                  or without parameter, display task without tags
                  -> ypodo untag linux 3 4 #delete the tag linux from the tasks identified by index equals 3 and 4
                  -> pypodo untag #display the todolist filtered on untagged tasks
        sort    : reorder the todolist in consecutives index
                  -> pypodo sort
        backup  : create a timestamped copy of the actual .todo file in a backupfolder
                  -> pypodo backup #create a timestamped copy of the actual .todo file in a backupfolder
        find    : filter the todolist on the [PARAMETER] (regex accepted)
                  -> pypodo filter "ta.*che" #filter the todolist on the regex ta.*che
        crypt   : crypt the todofile with the key [PARAMETER] or if not gived, with the property SYSTEM.KEY of the config file
                  the todofile crypted is saved in ~/.todo.crypt (can be used to crypt an existent no crypted todofile
                  after it, just replace todofile content with .todo.crypt content and add the SYSTEM.KEY in the config file)
                  -> pypodo crypt "admin" #create a ~/.todo.crypt crypted from todofile with the key "admin"
                  -> pypodo crypt #create a ~/.todo.crypt crypted from todofile with the key from config file
        decrypt : decrypt the todofile with the key [PARAMETER] or if not gived, with the property SYSTEM.KEY of the config file
                  the todofile decrypted is saved in ~/.todo.decrypt (can be used to decrypt an existent crypted todofile
                  after it, just replace todofile content with .todo.decrypt content and delete the SYSTEM.KEY in the config file)
                  -> pypodo decrypt "admin" #create a ~/.todo.decrypt crypted from todofile with the key "admin"
                  -> pypodo decrypt #create a ~/.todo.decrypt crypted from todofile with the key from config file
        """
    print(help_txt)


def listtask(openfile=open):
    """
    Print the todofile with filters or not
    """
    if check(openfile):

        empty = True
        with openfile(todofilefromconfig(), "r") as todofile:
            for line in todofile.readlines():
                # without filter -> we print all
                if len(sys.argv) == 2:
                    printlinetodo(line)
                    empty = False
                # with filter -> we check tag
                else:
                    display = True
                    for increment in range(2, len(sys.argv)):
                        tagtofilter = sys.argv[increment]
                        if not re.findall(
                            " #"
                            + re.escape(tagtofilter)
                            + REGEX_SPACE_OR_ENDLINE,
                            line.rstrip("\n"),
                        ):
                            display = False
                    # regex to search tags "#toto " or "#toto" at the end of the line
                    if display:
                        printlinetodo(line)
                        empty = False
        if empty:
            if len(sys.argv) >= 3:
                printwarning("the filtered todolist is empty")
            else:
                printwarning("the todolist is empty")


def add(openfile=open):
    """
    Add a task to the todofile
    """
    if check(openfile):
        if len(sys.argv) < 3:
            printerror(
                "1 or more parameter is needed for pypodo add - tasks"
            )
        else:
            with openfile(todofilefromconfig(), "r") as todofile:
                lines = todofile.readlines()
                # index calculation
                if len(lines) > 0:
                    last_line = lines[len(lines) - 1]
                    index = int(last_line.split()[0]) + 1
                else:
                    index = 1
            with openfile(todofilefromconfig(), "a") as todofile:
                # loop on the indexes
                for increment in range(2, len(sys.argv)):
                    task = sys.argv[increment]
                    # check format : words* #tag1 #tag2 : task at free format,
                    # tags in one word prefixed by #
                    if not re.findall(
                        "^([^#]|([^ ]#))*( #[^ #]+)*$", task
                    ):
                        printwarning(
                            "the task has not a valid format - "
                            + task
                        )
                    else:
                        # adding task to the todolist
                        todofile.write(str(index) + " " + task + "\n")
                        printinfo(
                            "task is added to the todolist - "
                            + str(index)
                            + " "
                            + task
                        )
                        index = index + 1


def delete(openfile=open):
    """
    Delete a task from the todofile
    """
    if check(openfile):
        if len(sys.argv) >= 3:
            # loop on the indexes
            for increment in range(2, len(sys.argv)):
                index = sys.argv[increment]
                # check the numeric format of the index

                if not re.findall(REGEX_INDEX, index):
                    printwarning(
                        "the index to delete is not in numeric format - "
                        + index
                    )
                else:
                    index_existant = False
                    with openfile(
                        todofilefromconfig(), "r"
                    ) as todofile:
                        lines = todofile.readlines()
                    with openfile(
                        todofilefromconfig(), "w"
                    ) as todofile:
                        for line in lines:
                            # if the current row doesn't contain the index it is kept
                            if not re.findall(
                                "^" + index + " ", line
                            ):
                                todofile.write(line)
                            # else it is deleted by not being copied
                            else:
                                printinfo(
                                    "task deleted from the todolist - "
                                    + line.rstrip("\n")
                                )
                                index_existant = True
                    if not index_existant:
                        printwarning(
                            "no task is deleted from the todolist, not existing index - "
                            + index
                        )
            if (
                read_config_boolean("FONCTIONAL", "autosort", "False")
                == "True"
            ):
                sys.argv = [sys.argv[0]]
                sort(openfile)
        else:
            printerror(
                "1 or more parameter is needed for pypodo del - indexes to delete in numeric format"
            )


def sort(openfile=open):
    """
    Reorder the todofile with consecutives indexes
    """
    if check(openfile):
        if len(sys.argv) > 2:
            printerror("0 parameter is needed for pypodo sort")
        else:
            empty = True
            index = 1
            with openfile(todofilefromconfig(), "r") as todofile:
                lines = todofile.readlines()
            with openfile(todofilefromconfig(), "w") as todofile:
                for line in lines:
                    # we replace the existing index by the current index that we increment
                    replaced = re.sub(
                        "^\\d+ ", str(index) + " ", line
                    )
                    index = index + 1
                    todofile.write(replaced)
                    empty = False
            if empty:
                printwarning("the todolist is empty - nothing to do")
            else:
                printinfo("the todolist is sorted")
                listtask(openfile)


def untag(openfile=open):
    """
    Untag tasks from the todofile
    """
    if check(openfile):
        if len(sys.argv) == 2:
            listnotag(openfile)
        elif len(sys.argv) >= 4:
            tagtodel = sys.argv[2]
            if not re.findall("^[^ #]+$", tagtodel):
                printerror(
                    "the tag has not a valid format - " + tagtodel
                )
            else:
                # loop on the indexes
                for increment in range(3, len(sys.argv)):
                    index = sys.argv[increment]
                    if not re.findall(REGEX_INDEX, index):
                        printwarning(
                            "the index to untag is not in numeric format - "
                            + index
                        )
                    else:
                        index_trouve = False
                        with openfile(
                            todofilefromconfig(), "r"
                        ) as todofile:
                            lines = todofile.readlines()
                        with openfile(
                            todofilefromconfig(), "w"
                        ) as todofile:
                            for line in lines:
                                if not re.findall(
                                    "^" + index + " ", line
                                ):
                                    todofile.write(line)
                                else:
                                    if re.findall(
                                        " #"
                                        + re.escape(tagtodel)
                                        + REGEX_SPACE_OR_ENDLINE,
                                        line.rstrip("\n"),
                                    ):
                                        newline = re.sub(
                                            "#"
                                            + re.escape(tagtodel)
                                            + REGEX_SPACE_OR_ENDLINE,
                                            "",
                                            line,
                                        ).rstrip("\n")
                                        todofile.write(
                                            newline.rstrip() + "\n"
                                        )
                                        printinfo(
                                            "tag deleted from the task of the todolist - "
                                            + line.rstrip("\n")
                                            + " -> "
                                            + newline
                                        )
                                    else:
                                        todofile.write(line)
                                        printwarning(
                                            "no tags is deleted from the todolist for the task - "
                                            + line.rstrip("\n")
                                        )
                                    index_trouve = True
                        if not index_trouve:
                            printwarning(
                                "no task with index - " + index
                            )
        else:
            printerror(
                "0,2 or more parameters is needed for pypodo untag : the tag to delete and"
                " the indexes of the task whose tags to delete - nothing to list task without tags"
            )


def tag(openfile=open):
    """
    Tag tasks from the todofile
    """
    if check(openfile):
        if len(sys.argv) == 2:
            listtag(openfile)
        elif len(sys.argv) >= 4:
            tagtoadd = sys.argv[2]
            if not re.findall("^[^ #]+$", tagtoadd):
                printerror(
                    "the tag has not a valid format - " + tagtoadd
                )
            else:
                # loop on the indexes
                for increment in range(3, len(sys.argv)):
                    index = sys.argv[increment]
                    if not re.findall(REGEX_INDEX, index):
                        printwarning(
                            "the index to tag is not in numeric format - "
                            + index
                        )
                    else:
                        index_trouve = False
                        with openfile(
                            todofilefromconfig(), "r"
                        ) as todofile:
                            lines = todofile.readlines()
                        with openfile(
                            todofilefromconfig(), "w"
                        ) as todofile:
                            for line in lines:
                                if not re.findall(
                                    "^" + index + " ", line
                                ):
                                    todofile.write(line)
                                else:
                                    newline = (
                                        line.rstrip("\n")
                                        + " #"
                                        + tagtoadd
                                    )
                                    todofile.write(newline + "\n")
                                    printinfo(
                                        "tag added to the task of the todolist - "
                                        + line.rstrip("\n")
                                        + " -> "
                                        + newline
                                    )
                                    index_trouve = True
                        if not index_trouve:
                            printwarning(
                                "no task with index - " + index
                            )
        else:
            printerror(
                "0,2 or more parameters is needed for pypodo tag : the tag to add and"
                " the indexes of the task whose tags to add - nothing to list tags of"
                " the todolist"
            )


def backup(openfile=open):
    """
    Backup the todofile
    """
    if check(openfile):
        if len(sys.argv) > 2:
            printerror("0 parameter is needed for pypodo backup")
        else:
            dir_exists = os.path.exists(todobackupfoderfromconfig())
            if not dir_exists:
                try:
                    os.makedirs(todobackupfoderfromconfig())
                except PermissionError:
                    printerror(
                        "permission error to create the backup folder : "
                        + todobackupfoderfromconfig()
                    )
                    sys.exit()
                printinfo("creating todolist backup folder")
            time_suffix = time.strftime("%Y%m%d%H%M%S")
            todo_backup_name = ".todo" + time_suffix
            backup_name = (
                todobackupfoderfromconfig() + todo_backup_name
            )
            try:
                copyfile(todofilefromconfig(), backup_name)
            except PermissionError:
                printerror(
                    "permission error to create the backup folder : "
                    + todobackupfoderfromconfig()
                )
                sys.exit()
            printinfo(
                "creating todolist backup - " + todo_backup_name
            )


def find(openfile=open):
    """
    Search with regex in the todofile
    """
    if check(openfile):
        if len(sys.argv) != 3:
            printerror("1 parameter is needed for pypodo find")
        else:
            empty = True
            with open(todofilefromconfig(), "r") as todofile:
                for line in todofile.readlines():
                    search = sys.argv[2]
                    if re.findall(search, line.rstrip("\n")):
                        printlinetodo(line)
                        empty = False
            if empty:
                printwarning("the filtered todolist is empty")


# secondary functions
def crypt(itself, cle):
    """
    Crypt the file in parameter to the standard systout
    """
    if itself:
        my_password = read_config("SYSTEM", "key", "")
    else:
        if not cle:
            my_password = read_config("SYSTEM", "key", "")
        else:
            my_password = cle

    if not my_password:
        if itself:
            printdebug("not crypt key")
        else:
            printerror("not crypt key")
    else:
        fernet_key = obtain_fernet_key(my_password)
        if not os.path.isfile(todofilefromconfig()):
            printerror("not .todo file")
        else:
            with open(todofilefromconfig(), "rb") as file:
                # read all file data
                file_data = file.read()
            # encrypt data
            encrypted_data = fernet_key.encrypt(file_data)
            printdebug("crypted file - " + str(encrypted_data))
            if itself:
                with open(todofilefromconfig(), "wb") as file:
                    file.write(encrypted_data)
            else:
                with open(STR_PATH_HOME__TODOCRYPT_, "wb") as file:
                    file.write(encrypted_data)
                    printinfo(
                        ".todo is crypted in "
                        + STR_PATH_HOME__TODOCRYPT_
                    )


def decrypt(itself, cle, openfile=open):
    """
    Crypt the file in parameter to the standard systout
    """
    if itself:
        my_password = read_config("SYSTEM", "key", "")
    else:
        if not cle:
            my_password = read_config("SYSTEM", "key", "")
        else:
            my_password = cle

    if not my_password:
        if itself:
            printdebug("not decrypt key")
        else:
            printerror("not decrypt key")
    else:
        fernet_key = obtain_fernet_key(my_password)
        if not itself and not os.path.isfile(todofilefromconfig()):
            printerror("not .todo file")
        else:
            if not os.path.isfile(todofilefromconfig()):
                printdebug("creating .todolist file")
                try:
                    openfile(todofilefromconfig(), "w")
                except FileNotFoundError:
                    printerror(
                        "the path "
                        + todofilefromconfig()
                        + " does not exist, correct it (in the .todo.rc file)"
                    )
                    sys.exit()
            else:
                if os.stat(todofilefromconfig()).st_size == 0:
                    printinfo("emtpy todofile")
                else:
                    with open(todofilefromconfig(), "rb") as file:
                        # read all file data
                        file_data = file.read()
                    # encrypt data
                    try:
                        printdebug("crypted file - " + str(file_data))
                        decrypted_data = fernet_key.decrypt(file_data)
                        printdebug(
                            "timestamp of crypted file - "
                            + str(
                                datetime.fromtimestamp(
                                    fernet_key.extract_timestamp(
                                        file_data
                                    )
                                )
                            )
                        )
                        if itself:
                            with open(
                                todofilefromconfig(), "wb"
                            ) as file:
                                file.write(decrypted_data)
                        else:
                            with open(
                                STR_PATH_HOME__TODODECRYPT_, "wb"
                            ) as file:
                                file.write(decrypted_data)
                                printinfo(
                                    ".todo is decrypted in "
                                    + STR_PATH_HOME__TODODECRYPT_
                                )
                    except InvalidToken:
                        printerror(
                            "can t decrypt todo, not crypted or bad password?"
                        )
                        sys.exit()


def obtain_fernet_key(my_password):
    """
    return fernet instance
    """
    return Fernet(generate_key_fernet(my_password))


def generate_key_fernet(my_password):
    """
    return fernet key b64 from every string (string->md5->64)
    """
    printdebug("fernet password - " + my_password)
    key = base64.b64encode(
        hashlib.md5(my_password.encode("utf-8"))
        .hexdigest()
        .encode("ascii")
    ).decode("ascii")
    printdebug("fernet key - " + key)
    return key


def listnotag(openfile=open):
    """
    Print the todofile filtered on tasks with not tags
    """
    printdebug("list tasks without tag")
    if check(openfile):
        if len(sys.argv) > 2:
            printerror("0 parameter is needed for pypodo listnotag")
        else:
            empty = True
            with openfile(todofilefromconfig(), "r") as todofile:
                for line in todofile.readlines():
                    if not " #" in line:
                        printlinetodo(line)
                        empty = False
            if empty:
                printwarning(
                    "the filtered todolist with no tag is empty"
                )


def listtag(openfile=open):
    """
    Print the tags of the todofile
    """
    printdebug("list uniq tags from task")
    if check(openfile):
        if len(sys.argv) > 2:
            printerror("0 parameter is needed for pypodo listtag")
        else:
            empty = True
            with openfile(todofilefromconfig(), "r") as todofile:
                my_list = []
                for line in todofile.readlines():
                    for part in line.split():
                        if part.startswith("#"):
                            if part in listalerttags():
                                part = colored(part, color_alert())
                            elif test_date(part[1:]) == "alert":
                                part = colored(part, color_alert())
                            elif test_date(part[1:]) == "warning":
                                part = colored(part, color_warning())
                            else:
                                part = colored(part, color_tag())
                            my_list.append(part)
                            empty = False
            if empty:
                printwarning("the list of todolist's tags is empty")
            else:
                print("\n".join(sorted(set(my_list))))


def check(openfile=open):
    """
    Check the toodofile
    """
    file_exists = os.path.isfile(todofilefromconfig())
    if file_exists:
        try:
            with openfile(todofilefromconfig(), "r") as todofile:
                error = False
                for line in todofile.readlines():
                    # verification regex, index + task + possible tags
                    if not re.findall(
                        "^\\d+ ([^#]|([^ ]#))*( #[^ #]+)*$",
                        line.rstrip("\n"),
                    ):
                        printwarning(
                            "this line has not a valid format in .todo - "
                            + line.rstrip("\n")
                        )
                        error = True
            if error:
                printerror("verify the .todo file. Is it encrypted?")
                sys.exit()
            return True
        except PermissionError:
            printerror(
                "permission error to open the todofile : "
                + todofilefromconfig()
            )
            sys.exit()

    printinfo("creating .todolist file")
    try:
        openfile(todofilefromconfig(), "a")
    except FileNotFoundError:
        printerror(
            "the path "
            + todofilefromconfig()
            + " does not exist, correct it (in the .todo.rc file)"
        )
        sys.exit()
    return True


# others functions
def todofilefromconfig():
    """
    Obtain path to todofile
    """
    return read_config(
        "SYSTEM", "todofile", str(Path.home()) + "/.todo"
    )


def todobackupfoderfromconfig():
    """
    Obtain path to todobackupfolder
    """
    return read_config(
        "SYSTEM",
        "todobackupfolder",
        str(Path.home()) + "/.todo_backup/",
    )


def periodalert():
    """
    Obtain period alert
    """
    return int(read_config_int("FONCTIONAL", "periodalert", "0"))


def periodwarning():
    """
    Obtain period warnning
    """
    return int(read_config_int("FONCTIONAL", "periodwarning", "7"))


def listalerttags():
    """
    Obtain list of alert tags
    """
    return read_config("FONCTIONAL", "alerttags", "#urgent").split(
        ","
    )


# print functions
def printlinetodo(line):
    """
    Display task with colors
    """
    task = colored(
        re.sub(" #.*", "", re.sub("^[^ ]+ ", "", line.rstrip("\n"))),
        color_task(),
    )
    index = colored(line.split(" ", 1)[0], color_index())
    tags = ""
    for part in line.split():
        if part.startswith("#"):
            if part in listalerttags():
                tags = tags + " " + (colored(part, color_alert()))
            elif test_date(part[1:]) == "alert":
                tags = tags + " " + (colored(part, color_alert()))
            elif test_date(part[1:]) == "warning":
                tags = tags + " " + (colored(part, color_warning()))
            else:
                tags = tags + " " + (colored(part, color_tag()))
    print(index + " " + task + tags)


def printdebug(text):
    """
    Color and key word debug for print
    """
    if read_config_level("SYSTEM", "messagelevel", "info") == "debug":
        print(colored("debug   : " + text, color_debug()))


def printinfo(text):
    """
    Color and key word info for print
    """
    if (
        read_config_level("SYSTEM", "messagelevel", "info") == "info"
        or read_config_level("SYSTEM", "messagelevel", "info")
        == "debug"
    ):
        print(colored("info    : " + text, color_info()))


def printwarning(text):
    """
    Color and key word warning for print
    """
    if (
        read_config_level("SYSTEM", "messagelevel", "info") == "info"
        or read_config_level("SYSTEM", "messagelevel", "info")
        == "debug"
        or read_config_level("SYSTEM", "messagelevel", "info")
        == "warning"
    ):
        print(colored("warning : " + text, color_warning()))


def printerror(text):
    """
    Color and key word error for print
    """
    print(colored("error   : " + text, color_alert()))


def color_debug():
    """
    Color for info
    """
    return read_config_color("COLOR", "debug", "grey")


def color_info():
    """
    Color for info
    """
    return read_config_color("COLOR", "info", "green")


def color_task():
    """
    Color for task
    """
    return read_config_color("COLOR", "task", "green")


def color_index():
    """
    Color for index
    """
    return read_config_color("COLOR", "index", "yellow")


def color_tag():
    """
    Color for tag
    """
    return read_config_color("COLOR", "tag", "cyan")


def color_warning():
    """
    Color for warning
    """
    return read_config_color("COLOR", "warning", "yellow")


def color_alert():
    """
    Color for alert
    """
    return read_config_color("COLOR", "alert", "red")


# config functions
def read_config(section, cle, defaut, openfile=open):
    """
    Read the config file
    """
    config = configparser.ConfigParser()
    try:
        try:
            try:
                openfile(STR_PATH_HOME__TODORC_, "r")
            except PermissionError:
                print(
                    colored(
                        "error   : permission error to open the ~/todo.rc file",
                        "red",
                    )
                )
                sys.exit()
        except FileNotFoundError:
            openfile(STR_PATH_HOME__TODORC_, "w")

        config.read(STR_PATH_HOME__TODORC_)
        return config[section][cle]
    except (configparser.MissingSectionHeaderError, KeyError):
        return defaut


def read_config_color(section, cle, defaut):
    """
    Read the config file for color
    """
    color = read_config(section, cle, defaut)
    if color not in [
        "grey",
        "red",
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
        "white",
    ]:
        return defaut
    return color


def read_config_int(section, cle, defaut):
    """
    Read the config file for natural number
    """
    number = read_config(section, cle, defaut)
    if not number.isdigit():
        return defaut
    return number


def read_config_boolean(section, cle, defaut):
    """
    Read the config file for color
    """
    myboool = read_config(section, cle, defaut)
    if myboool not in ["True", "False"]:
        return defaut
    return myboool


def read_config_level(section, cle, defaut):
    """
    Read the config file for level
    """
    level = read_config(section, cle, defaut)
    if level not in ["warning", "info", "error", "debug"]:
        return defaut
    return level


def read_config_date_format(section, cle, defaut):
    """
    Read the config file for date format
    """
    return read_config(section, cle, defaut)


def test_date(datetime_str):
    """
    Comparare date en return alert state
    """
    try:
        datetime_object = datetime.strptime(
            datetime_str,
            read_config_date_format("SYSTEM", "formatdate", "%Y%m%d"),
        ).date()
    except ValueError:
        return "ok"
    if (date.today() - datetime_object).days >= -1 * periodalert():
        return "alert"
    if (date.today() - datetime_object).days >= -1 * periodwarning():
        return "warning"
    return "ok"



"""
Pypodo Tests
"""

import re
import os
import sys
import unittest
from io import StringIO
from pathlib import Path
from unittest.mock import MagicMock, mock_open, patch

from freezegun import freeze_time

from pypodo.__pypodo__ import (
    add,
    backup,
    delete,
    find,
    helppypodo,
    listnotag,
    listtag,
    listtask,
    pypodo,
    read_config,
    read_config_boolean,
    read_config_color,
    read_config_int,
    read_config_level,
    sort,
    tag,
    test_date,
    untag,
    generate_key_fernet,
)


class AnyStringWith(str):
    """
    Utilitary Class for contains string
    """

    def __eq__(self, other):
        """
        Test if string is in other
        """
        return self in other


sys.modules["shutil"] = MagicMock()

STR_PATH_HOME__TODO_ = str(Path.home()) + "/.todo"
STR_PATH_HOME__TODORC_ = str(Path.home()) + "/.todo.rc"
STR_PATH_HOME__TODO_BACKUP_FOLDER_ = (
    str(Path.home()) + "/.todo_backup/"
)


class TestMethodsEntryPoint(unittest.TestCase):
    """
    Class Test for entrypoints
    """

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[SYSTEM]\nmessagelevel = debug",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_crypt(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_crypt
        """
        with patch.object(sys, "argv", [pypodo, "crypt"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : not crypt key",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[SYSTEM]\nmessagelevel = debug",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_decrypt(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        def test_pypodo_entry_point_decrypt
        """
        with patch.object(sys, "argv", [pypodo, "decrypt"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : not decrypt key",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_list(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_list
        """
        with patch.object(sys, "argv", [pypodo, "list"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the todolist is empty",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_add(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_add
        """
        with patch.object(sys, "argv", [pypodo, "add"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 1 or more parameter is needed for pypodo add - tasks",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_del(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_del
        """
        with patch.object(sys, "argv", [pypodo, "del"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 1 or more parameter is needed for pypodo del "
                "- indexes to delete in numeric format",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_sort(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_sort
        """
        with patch.object(sys, "argv", [pypodo, "sort"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the todolist is empty - nothing to do",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_untag(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_untag
        """
        with patch.object(sys, "argv", [pypodo, "untag", "tag"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0,2 or more parameters is needed for pypodo untag :"
                " the tag to delete and the indexes of the task whose tags"
                " to delete - nothing to list task without tags",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_tag(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_tag
        """
        with patch.object(sys, "argv", [pypodo, "tag", "tag"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0,2 or more parameters is needed for pypodo tag"
                " : the tag to add and the indexes of the task whose tags"
                " to add - nothing to list tags of the todolist",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_find(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_find
        """
        with patch.object(sys, "argv", [pypodo, "find"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 1 parameter is needed for pypodo find",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_entry_point_backup(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_pypodo_entry_point_backup
        """
        with patch.object(sys, "argv", [pypodo, "backup", "bak"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0 parameter is needed for pypodo backup",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    @patch("builtins.print", autospec=True, side_effect=print)
    def test_pypodo_entry_point_help(
        self, mock_print, mock_open_file, mock_isfile, mock_stdout
    ):
        """
        test_pypodo_entry_point_help
        """
        with patch.object(sys, "argv", [pypodo, "help"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            mock_print.assert_called_with(AnyStringWith("SYNOPSIS"))

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    @patch("builtins.print", autospec=True, side_effect=print)
    def test_pypodo_entry_point_bad_param(
        self, mock_print, mock_open_file, mock_isfile, mock_stdout
    ):
        """
        test_pypodo_entry_point_bad_param
        """
        with patch.object(sys, "argv", [pypodo, "param"]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            mock_print.assert_called_with(AnyStringWith("SYNOPSIS"))

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    @patch("builtins.print", autospec=True, side_effect=print)
    def test_pypodo_entry_point_no_param(
        self, mock_print, mock_open_file, mock_isfile, mock_stdout
    ):
        """
        def test_pypodo_entry_point_no_param
        """
        with patch.object(sys, "argv", [pypodo]):
            mock_isfile.return_value = True
            pypodo(mock_open_file)
            mock_print.assert_called_with(AnyStringWith("SYNOPSIS"))


class TestMethodsErrors(unittest.TestCase):
    """
    Class Test for Errors
    """

    # errors
    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_find_with_no_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_find_with_no_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, find]):
            mock_isfile.return_value = True
            find(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 1 parameter is needed for pypodo find",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_find_with_two_parameters_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_find_with_two_parameters_return_error
        """
        with patch.object(sys, "argv", [pypodo, find, "one", "two"]):
            mock_isfile.return_value = True
            find(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 1 parameter is needed for pypodo find",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_add_with_no_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_add_with_no_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, add]):
            mock_isfile.return_value = True
            add(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 1 or more parameter is needed for pypodo add"
                " - tasks",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_del_with_no_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_del_with_no_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, delete]):
            mock_isfile.return_value = True
            delete(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 1 or more parameter is needed for pypodo del"
                " - indexes to delete in numeric format",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_sort_with_one_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_sort_with_one_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, sort, "one"]):
            mock_isfile.return_value = True
            sort(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0 parameter is needed for pypodo sort",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_sort_with_two_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_sort_with_two_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, sort, "one", "two"]):
            mock_isfile.return_value = True
            sort(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0 parameter is needed for pypodo sort",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_backup_with_one_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_backup_with_one_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, backup, "one"]):
            mock_isfile.return_value = True
            backup(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0 parameter is needed for pypodo backup",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_backup_with_two_parameters_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_backup_with_two_parameters_return_error
        """
        with patch.object(
            sys, "argv", [pypodo, backup, "one", "two"]
        ):
            mock_isfile.return_value = True
            backup(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0 parameter is needed for pypodo backup",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_with_one_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_with_one_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, tag, "one"]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0,2 or more parameters is needed for"
                " pypodo tag : the tag to add and the indexes of"
                " the task whose tags to add - nothing to list tags"
                " of the todolist",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_with_one_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_with_one_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, untag, "param"]):
            mock_isfile.return_value = True
            untag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0,2 or more parameters is needed for"
                " pypodo untag : the tag to delete and the indexes"
                " of the task whose tags to delete - nothing to"
                " list task without tags",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_notag_with_no_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_notag_with_no_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, untag, "param"]):
            mock_isfile.return_value = True
            listnotag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0 parameter is needed for pypodo listnotag",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_listtag_with_no_parameter_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_listtag_with_no_parameter_return_error
        """
        with patch.object(sys, "argv", [pypodo, untag, "param"]):
            mock_isfile.return_value = True
            listtag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : 0 parameter is needed for pypodo listtag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_beginning_with_hash_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_beginning_with_hash_return_error
        """
        with patch.object(sys, "argv", [pypodo, tag, "#badtag", "1"]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : the tag has not a valid format - #badtag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_containing_hash_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_containing_hash_return_error
        """
        with patch.object(sys, "argv", [pypodo, tag, "bad#tag", "1"]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : the tag has not a valid format - bad#tag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_finishing_with_hash_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_finishing_with_hash_return_error
        """
        with patch.object(sys, "argv", [pypodo, tag, "badtag#", "1"]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : the tag has not a valid format - badtag#",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_beginning_with_hash_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_beginning_with_hash_return_error
        """
        with patch.object(
            sys, "argv", [pypodo, untag, "#badtag", "1"]
        ):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : the tag has not a valid format - #badtag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_containing_hash_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_containing_hash_return_error
        """
        with patch.object(
            sys, "argv", [pypodo, untag, "bad#tag", "1"]
        ):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : the tag has not a valid format - bad#tag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_finishing_with_hash_return_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_finishing_with_hash_return_error
        """
        with patch.object(
            sys, "argv", [pypodo, untag, "badtag#", "1"]
        ):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "error   : the tag has not a valid format - badtag#",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 my task #te#st\na other task",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_sort_with_invalid_todo_return_warning_and_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_sort_with_invalid_todo_return_warning_and_error
        """
        with patch.object(sys, "argv", [pypodo, listtask]):
            mock_isfile.return_value = True
            try:
                error = False
                listtask(mock_open_file)
            except SystemExit:
                self.assertEqual(
                    escape_ansi(mock_print.getvalue().rstrip("\n")),
                    "warning : this line has not a valid format in .todo"
                    " - 1 my task #te#st\nwarning : this line has not a"
                    " valid format in .todo - a other task\nerror   :"
                    " verify the .todo file. Is it encrypted?",
                )
                error = True
            self.assertEqual(error, True)

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 my task te#st\na other task",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_sort_with_other_invalid_todo_return_warning_and_error(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_sort_with_other_invalid_todo_return_warning_and_error
        """
        with patch.object(sys, "argv", [pypodo, listtask]):
            mock_isfile.return_value = True
            try:
                error = False
                listtask(mock_open_file)
            except SystemExit:
                self.assertEqual(
                    escape_ansi(mock_print.getvalue().rstrip("\n")),
                    "warning : this line has not a"
                    " valid format in .todo - a other task\nerror   :"
                    " verify the .todo file. Is it encrypted?",
                )
                error = True
            self.assertEqual(error, True)


class TestMethodsWarnings(unittest.TestCase):
    """
    Class Test for Warnings
    """

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[SYSTEM]\nmessagelevel = debug\ntodofile = /tmp/.todo",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_crypt_with_not_todofile(
        self, mock_print, mock_open_file
    ):
        """
        test_pypodo_crypt_with_not_todofile
        """
        with patch.object(sys, "argv", [pypodo, "crypt", "password"]):
            if os.path.isfile("/tmp/.todo"):
                os.remove("/tmp/.todo")
            pypodo()
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "debug   : fernet password - password"
                "\ndebug   : fernet key - NWY0ZGNjM2I1YWE3NjVkNjFkODMyN2RlYjg4MmNmOTk="
                "\nerror   : not .todo file",
            )

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[SYSTEM]\nmessagelevel = debug\ntodofile = /tmp/.todo\nkey = password",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_list_with_key_and_not_todofile(
        self, mock_print, mock_open_file
    ):
        """
        test_pypodo_list_with_key_and_not_todofile
        """
        with patch.object(sys, "argv", [pypodo, "list"]):
            if os.path.isfile("/tmp/.todo"):
                os.remove("/tmp/.todo")
            try:
                error = False
                pypodo()
            # FIXME ResourceWarning: unclosed file patch for error in test
            except TypeError:
                self.assertEqual(
                    escape_ansi(mock_print.getvalue().rstrip("\n")),
                    "debug   : fernet password - password"
                    "\ndebug   : fernet key - NWY0ZGNjM2I1YWE3NjVkNjFkODMyN2RlYjg4MmNmOTk="
                    "\ndebug   : creating .todolist file\ninfo    : creating .todolist file"
                    "\nwarning : the todolist is empty\ndebug   : fernet password - password"
                    "\ndebug   : fernet key - NWY0ZGNjM2I1YWE3NjVkNjFkODMyN2RlYjg4MmNmOTk=",
                )
                error = True
            self.assertEqual(error, True)

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[SYSTEM]\nmessagelevel = debug\ntodofile = /tmp/.todo",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_pypodo_decrypt_with_not_todofile(
        self, mock_print, mock_open_file
    ):
        """
        test_pypodo_decrypt_with_not_todofile
        """
        with patch.object(
            sys, "argv", [pypodo, "decrypt", "password"]
        ):
            if os.path.isfile("/tmp/.todo"):
                os.remove("/tmp/.todo")
            pypodo()
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "debug   : fernet password - password"
                "\ndebug   : fernet key - NWY0ZGNjM2I1YWE3NjVkNjFkODMyN2RlYjg4MmNmOTk="
                "\nerror   : not .todo file",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_sort_with_empty_todofile(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_sort_with_empty_todofile
        """
        with patch.object(sys, "argv", [pypodo, sort]):
            mock_isfile.return_value = True
            sort(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the todolist is empty - nothing to do",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_with_no_numeric_value_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_with_no_numeric_value_return_warning
        """
        with patch.object(sys, "argv", [pypodo, tag, "param", "a"]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the index to tag is not in numeric format - a",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_with_no_numeric_value_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_with_no_numeric_value_return_warning
        """
        with patch.object(sys, "argv", [pypodo, untag, "param", "a"]):
            mock_isfile.return_value = True
            untag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the index to untag is not in numeric format - a",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_delete_with_no_numeric_value_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_delete_with_no_numeric_value_return_warning
        """
        with patch.object(sys, "argv", [pypodo, delete, "a"]):
            mock_isfile.return_value = True
            delete(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the index to delete is not in numeric format - a",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_delete_with_inexistant_numeric_value_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_delete_with_inexistant_numeric_value_return_warning
        """
        with patch.object(sys, "argv", [pypodo, delete, "2"]):
            mock_isfile.return_value = True
            delete(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : no task is deleted from the todolist,"
                " not existing index - 2",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_with_inexistant_numeric_value_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_with_inexistant_numeric_value_return_warning
        """
        with patch.object(sys, "argv", [pypodo, tag, "tag", "2"]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : no task with index - 2",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_with_inexistant_numeric_value_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_with_inexistant_numeric_value_return_warning
        """
        with patch.object(sys, "argv", [pypodo, untag, "tag", "2"]):
            mock_isfile.return_value = True
            untag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : no task with index - 2",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_listtag_with_empty_result_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_listtag_with_empty_result_return_warning
        """
        with patch.object(sys, "argv", [pypodo, tag]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the list of todolist's tags is empty",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #tag",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_listnottag_with_empty_result_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_listnottag_with_empty_result_return_warning
        """
        with patch.object(sys, "argv", [pypodo, untag]):
            mock_isfile.return_value = True
            untag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the filtered todolist with no tag is empty",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_with_empty_todofile_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_with_empty_todofile_return_warning
        """
        with patch.object(sys, "argv", [pypodo, listtask]):
            mock_isfile.return_value = True
            listtask(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the todolist is empty",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #tag1\n2 task #tag2\n3 task #tag\n4 task",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_with_filter_and_empty_result_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_with_filter_and_empty_result_return_warning
        """
        with patch.object(
            sys, "argv", [pypodo, listtask, "tag1", "tag2"]
        ):
            mock_isfile.return_value = True
            listtask(mock_open_file)
            # FIXME how test?
            # mock_open.assert_called_with(STR_PATH_HOME__TODO_, 'r')
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the filtered todolist is empty",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_add_task_begining_with_hash_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_add_task_begining_with_hash_return_warning
        """
        with patch.object(sys, "argv", [pypodo, add, "#task"]):
            mock_isfile.return_value = True
            add(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the task has not a valid format - #task",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_add_task_begining_with_hash_and_with_tag_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_add_task_begining_with_hash_and_with_tag_return_warning
        """
        with patch.object(sys, "argv", [pypodo, add, "#task #task"]):
            mock_isfile.return_value = True
            add(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the task has not a valid format - #task #task",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_add_task_with_empty_hash_and_with_tag_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_add_task_with_empty_hash_and_with_tag_return_warning
        """
        with patch.object(sys, "argv", [pypodo, add, "task # #task"]):
            mock_isfile.return_value = True
            add(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the task has not a valid format - task # #task",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_add_task_containing_hash_and_return_no_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        def test_add_task_containing_hash_and_return_no_warning
        """
        with patch.object(
            sys, "argv", [pypodo, add, "task#task #task"]
        ):
            mock_isfile.return_value = True
            add(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : task is added to the todolist - 1 task#task #task",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_find_if_empty_result_return_warning(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_find_if_empty_result_return_warning
        """
        with patch.object(sys, "argv", [pypodo, find, "string"]):
            mock_isfile.return_value = True
            find(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "warning : the filtered todolist is empty",
            )


class TestMethodsTools(unittest.TestCase):
    """
    Class Test for Tools
    """

    @freeze_time("2020-10-14")
    @patch(
        "builtins.open",
        new_callable=mock_open,
    )
    def test_date_without_onfig(self, mock_open_file):
        """
        test_date_without_onfig
        """
        self.assertEqual("ok", test_date("20210101"))
        self.assertEqual("alert", test_date("20201009"))
        self.assertEqual("warning", test_date("20201020"))

    @freeze_time("2020-10-14")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[SYSTEM]\nformatdate = %%Y%%d%%m",
    )
    def test_date_with_config(self, mock_open_file):
        """
        test_date_with_config
        """
        self.assertEqual("ok", test_date("20210101"))
        self.assertEqual("alert", test_date("20200910"))
        self.assertEqual("warning", test_date("20202010"))

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[COLOR]\nalert =red\nwarning=blue\ninfo=bad",
    )
    def test_color_with_config(self, mock_open_file):
        """
        test_color_with_config
        """
        # read_config
        self.assertEqual("red", read_config("COLOR", "error", "red"))
        self.assertEqual("red", read_config("COLOR", "alert", "red"))
        self.assertEqual(
            "blue", read_config("COLOR", "warning", "yellow")
        )
        self.assertEqual("bad", read_config("COLOR", "info", "green"))

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[COLOR]\nalert =red\nwarning=blue\ninfo=bad",
    )
    def test_color_with_other_config(self, mock_open_file):
        """
        test_color_with_other_config
        """
        # read_config_color
        self.assertEqual(
            "red", read_config_color("COLOR", "error", "red")
        )
        self.assertEqual(
            "red", read_config_color("COLOR", "alert", "red")
        )
        self.assertEqual(
            "blue", read_config_color("COLOR", "warning", "yellow")
        )
        self.assertEqual(
            "green", read_config_color("COLOR", "info", "green")
        )

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[FONCTIONAL]\nperiodwarning = bad\nperiodalert = 1\nperiodinfo = -1",
    )
    def test_period_without_config(self, mock_open_file):
        """
        test_period_without_config
        """
        # read_config_int
        self.assertEqual(
            "1", read_config_int("FONCTIONAL", "periodalert", "0")
        )
        self.assertEqual(
            "7", read_config_int("FONCTIONAL", "periodwarning", "7")
        )
        self.assertEqual(
            "10", read_config_int("FONCTIONAL", "periodinfo", "10")
        )

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[SYSTEM]\nlevel = error\nlevel2 = warning\nlevel3 = bad",
    )
    def test_period_with_config(self, mock_open_file):
        """
        test_period_with_config
        """
        # read_config_level
        self.assertEqual(
            "error", read_config_level("SYSTEM", "level", "error")
        )
        self.assertEqual(
            "warning", read_config_level("SYSTEM", "level2", "error")
        )
        self.assertEqual(
            "error", read_config_level("SYSTEM", "level3", "error")
        )

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="[FONCTIONAL]\nmybool = False\nmybool2 = bad",
    )
    def test_boolean_with_config(self, mock_open_file):
        """
        test_read_config_boolean
        """
        # read_config_boolean
        self.assertEqual(
            "False",
            read_config_boolean("FONCTIONAL", "mybool", "True"),
        )
        self.assertEqual(
            "True",
            read_config_boolean("FONCTIONAL", "mybool2", "True"),
        )


class TestMethodsOthers(unittest.TestCase):
    """
    Class Test for other
    """

    @patch("time.strftime")
    @patch("pypodo.__pypodo__.copyfile")
    @patch("os.path.isfile")
    @patch("os.makedirs")
    @patch("os.path.exists", autospec=True)
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_backup_create_file_with_good_sufix_and_good_content(
        self,
        mock_print,
        mock_open_file,
        mock_path_exists,
        mock_makedirs,
        mock_isfile,
        mock_copyfile,
        mock_strftime,
    ):
        """
        test_backup_create_file_with_good_sufix_and_good_content
        """
        with patch.object(sys, "argv", [pypodo, backup]):
            mock_isfile.return_value = True
            mock_path_exists.return_value = False
            mock_strftime.return_value = "_sufix_date"
            backup(mock_open_file)
            mock_path_exists.assert_called_with(
                STR_PATH_HOME__TODO_BACKUP_FOLDER_
            )
            mock_makedirs.assert_called_with(
                STR_PATH_HOME__TODO_BACKUP_FOLDER_
            )
            mock_copyfile.assert_called_with(
                STR_PATH_HOME__TODO_,
                STR_PATH_HOME__TODO_BACKUP_FOLDER_
                + ".todo_sufix_date",
            )
            self.assertIn(
                "info    : creating todolist backup folder",
                escape_ansi(mock_print.getvalue().rstrip("\n")),
            )
            self.assertIn(
                "creating todolist backup - .todo_sufix_date",
                escape_ansi(mock_print.getvalue().rstrip("\n")),
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="2 task one #test\n4 task two",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_sort_reorganize_index(
        self, mock_sysout, mock_open_file, mock_isfile
    ):
        """
        test_sort_reorganize_index
        """
        with patch.object(sys, "argv", [pypodo, sort]):
            mock_isfile.return_value = True
            sort(mock_open_file)
            mock_open_file().write.assert_called_with("2 task two")

    @patch("os.path.isfile")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="1 task"
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_print_one_task_if_one_task_in_todofile(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_print_one_task_if_one_task_in_todofile
        """
        with patch.object(sys, "argv", [pypodo, listtask]):
            mock_isfile.return_value = True
            listtask(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "1 task",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 t#urgentask# #tag1 #tag2",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_print_one_task_with_tags_if_one_task_in_todofile(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_print_one_task_with_tags_if_one_task_in_todofile
        """
        with patch.object(sys, "argv", [pypodo, listtask]):
            mock_isfile.return_value = True
            listtask(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "1 t#urgentask# #tag1 #tag2",
            )

    @freeze_time("2020-10-14")
    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #urgent #20201010 #20201015 #20201022 #tag",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_print_one_task_with_specific_tags_if_one_task_in_todofile(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_print_one_task_with_specific_tags_if_one_task_in_todofile
        """
        with patch.object(sys, "argv", [pypodo, listtask]):
            mock_isfile.return_value = True
            listtask(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "1 task #urgent #20201010 #20201015 #20201022 #tag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task1 #tag\n2 task2 #tag2\n3 task3#tag #tag3",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_with_filter_return_result(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_with_filter_return_result
        """
        with patch.object(sys, "argv", [pypodo, listtask, "tag"]):
            mock_isfile.return_value = True
            listtask(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "1 task1 #tag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #tag1\n2 task2 tache #tag2\n3 task3 #tag1 #tag2\n4 ma 4 task4 #tag3",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_with_double_filter_return_result_with_combinate(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_with_double_filter_return_result_with_combinate
        """
        with patch.object(
            sys, "argv", [pypodo, listtask, "tag1", "tag2"]
        ):
            mock_isfile.return_value = True
            listtask(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "3 task3 #tag1 #tag2",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_add_multi_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_add_multi_task
        """
        with patch.object(
            sys, "argv", [pypodo, add, "task1", "task2"]
        ):
            mock_isfile.return_value = True
            add(mock_open_file)
            mock_open_file().write.assert_called_with("2 task2\n")
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : task is added to the todolist - 1 task1\n"
                "info    : task is added to the todolist - 2 task2",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task1\n4 task2",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_add_task_with_no_empty_todolist(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_add_task_with_no_empty_todolist
        """
        with patch.object(sys, "argv", [pypodo, add, "task3"]):
            mock_isfile.return_value = True
            add(mock_open_file)
            # mock_open.assert_called_with(STR_PATH_HOME__TODO_, 'a')
            mock_open_file().write.assert_called_once_with(
                "5 task3\n"
            )
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : task is added to the todolist - 5 task3",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task1 #tag\n2 task2\n4 task3",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_delete_one_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_delete_one_task
        """
        with patch.object(sys, "argv", [pypodo, delete, "2"]):
            mock_isfile.return_value = True
            delete(mock_open_file)
            mock_open_file().write.assert_called_with("4 task3")
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : task deleted from the todolist - 2 task2",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task1 #tag\n2 task2\n4 task3",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_delete_multi_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_delete_multi_task
        """
        with patch.object(sys, "argv", [pypodo, delete, "2", "1"]):
            mock_isfile.return_value = True
            delete(mock_open_file)
            mock_open_file().write.assert_called_with("4 task3")
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : task deleted from the todolist - 2 task2"
                "\ninfo    : task deleted from the todolist - 1 task1 #tag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #tag2\n2 task2 #tag #tag3",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_one_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_one_task
        """
        with patch.object(sys, "argv", [pypodo, untag, "tag", "2"]):
            mock_isfile.return_value = True
            untag(mock_open_file)
            mock_open_file().write.assert_called_with(
                "2 task2 #tag3\n"
            )
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : tag deleted from the task of the todolist"
                " - 2 task2 #tag #tag3 -> 2 task2 #tag3",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task#tag #tag2\n2 task#tag2 #tag #tag3",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_other_one_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_one_task
        """
        with patch.object(sys, "argv", [pypodo, untag, "tag", "2"]):
            mock_isfile.return_value = True
            untag(mock_open_file)
            mock_open_file().write.assert_called_with(
                "2 task#tag2 #tag3\n"
            )
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : tag deleted from the task of the todolist"
                " - 2 task#tag2 #tag #tag3 -> 2 task#tag2 #tag3",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 ma tache #test #toto\n2 ma seconde tache #tost"
        " #titi\n3 ma seconde tache #test #titi",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_untag_multi_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_untag_multi_task
        """
        with patch.object(
            sys, "argv", [pypodo, untag, "test", "1", "2", "3"]
        ):
            mock_isfile.return_value = True
            untag(mock_open_file)
            # mock_open.assert_called_with(STR_PATH_HOME__TODO_, 'w')
            mock_open_file().write.assert_called_with(
                "3 ma seconde tache #titi\n"
            )
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : tag deleted from the task of the todolist"
                " - 1 ma tache #test #toto -> 1 ma tache #toto\n"
                "warning : no tags is deleted from the todolist for the task"
                " - 2 ma seconde tache #tost #titi\ninfo    : tag deleted"
                " from the task of the todolist - 3 ma seconde tache"
                " #test #titi -> 3 ma seconde tache #titi",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task1\n2 task2 #tag1",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_one_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_one_task
        """
        with patch.object(sys, "argv", [pypodo, tag, "tag2", "2"]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            mock_open_file().write.assert_called_with(
                "2 task2 #tag1 #tag2\n"
            )
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : tag added to the task of the todolist"
                " - 2 task2 #tag1 -> 2 task2 #tag1 #tag2",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task1\n2 task2 #tag1\n3 task3 #tag2",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_tag_multi_task(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_tag_multi_task
        """
        with patch.object(
            sys, "argv", [pypodo, tag, "tag3", "2", "3"]
        ):
            mock_isfile.return_value = True
            tag(mock_open_file)
            mock_open_file().write.assert_called_with(
                "3 task3 #tag2 #tag3\n"
            )
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : tag added to the task of the todolist"
                " - 2 task2 #tag1 -> 2 task2 #tag1 #tag3\n"
                "info    : tag added to the task of the todolist"
                " - 3 task3 #tag2 -> 3 task3 #tag2 #tag3",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task1 #tag\n2 task2\n3 task3\n",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_list_notag_return_result(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_list_notag_return_result
        """
        with patch.object(sys, "argv", [pypodo, untag]):
            mock_isfile.return_value = True
            untag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "2 task2\n3 task3",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 ta#urgentsk #tag\n2 task2 #tag\n3 task3 #tag2\n4 task4",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_listtag_return_other_result(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_listtag_return_result
        """
        with patch.object(sys, "argv", [pypodo, tag]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "#tag\n" + "#tag2",
            )

    @freeze_time("2020-10-14")
    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #urgent\n2 task2 #20190101"
        "\n3 task3 #20201020\n4 task4 #tag",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_listtag_with_specific_tags_return_result(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_listtag_with_specific_tags_return_result
        """
        with patch.object(sys, "argv", [pypodo, tag]):
            mock_isfile.return_value = True
            tag(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "#20190101\n#urgent\n#20201020\n#tag",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #tag\n2 specif_task #tag2\n3 specif",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_find_one_task_without_regex(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_find_one_task_without_regex
        """
        with patch.object(sys, "argv", [pypodo, find, "specif_task"]):
            mock_isfile.return_value = True
            find(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "2 specif_task #tag2",
            )

    @patch("os.path.isfile")
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="1 task #lix\n2 task2 #tag\n3 task3 #linux",
    )
    @patch("sys.stdout", new_callable=StringIO)
    def test_find_with_regex(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_find_with_regex
        """
        with patch.object(sys, "argv", [pypodo, find, "li.+x"]):
            mock_isfile.return_value = True
            find(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "3 task3 #linux",
            )

    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open)
    @patch("sys.stdout", new_callable=StringIO)
    def test_find_with_regex_no_result(
        self, mock_print, mock_open_file, mock_isfile
    ):
        """
        test_find_with_regex_no_result
        """
        mock_isfile.return_value = False
        with patch.object(sys, "argv", [pypodo, find, "li.+x"]):
            find(mock_open_file)
            self.assertEqual(
                escape_ansi(mock_print.getvalue().rstrip("\n")),
                "info    : creating .todolist file\nwarning :"
                " the filtered todolist is empty",
            )

    @patch("sys.stdout", new_callable=StringIO)
    @patch("builtins.print", autospec=True, side_effect=print)
    def test_help_contains_synopsis(self, mock_print, mock_stdout):
        """
        test_help_contains_synopsis
        """
        helppypodo()
        mock_print.assert_called_with(AnyStringWith("SYNOPSIS"))

    def test_generate_key_fernet(self):
        """
        test key generators
        """
        self.assertEqual(
            generate_key_fernet("password"),
            "NWY0ZGNjM2I1YWE3NjVkNjFkODMyN2RlYjg4MmNmOTk=",
        )


def escape_ansi(line):
    """
    escape_colors_from_string
    """
    ansi_escape = re.compile(
        r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]"
    )
    return ansi_escape.sub("", line)


if __name__ == "__main__":
    unittest.main()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pypodo",
    version="3.0.2",
    description="pypodo is a todolist tool which works with a .todo file at the root of the home directory. It has a mecanism of indexes and tags.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thib1984/pypodo",
    author="thib1984",
    author_email="thibault.garcon@gmail.com",
    license="mit",
    packages=["pypodo"],
    install_requires=["setuptools", "termcolor", "cryptography"],
    zip_safe=False,
    entry_points={
        "console_scripts": ["pypodo=pypodo.__pypodo__:pypodo"],
    },
)
